import asyncio
import sys

delay = 1

def zag_sampler_loop(lc):
    loop, callback = lc

    loop.call_later(delay, callback, lc)
    sys.stdout.write(".")
    sys.stdout.flush()


if __name__ == "__main__":
    loop = asyncio.get_event_loop()

    callback = zag_sampler_loop
    lc = (loop, callback)
    loop.call_later(delay, callback, lc)

    loop.run_forever()
    loop.close()



#!/usr/bin/env python

import psutil
import logging
import os
import time
import zaggregator.utils as utils

DEFAULT_INTERVAL  = 1.0
interpreters = "python", "perl", "bash", "tcsh", "zsh",

class ProcessMirror:
    """
        Class provides cachable and transparent copy of psutil.Process
        with some fields overwritten to avoid direct calling of
        psutil.Process methods, as they are inconsistent for dead processes

    """
    def __init__(self, proc: psutil.Process, proctable):
        """
            Class constructor
        """

        _dead = False
        self._children = list()
        self._parent = 0
        parent = None

        self._proc, self._pt = proc, proctable
        try:
            self._pgid = os.getpgid(proc.pid)
        except:
            self._pgid = 0
            _dead = True

        if not _dead:
            parent = proc.parent()

        if parent:
            self._parent = parent.pid

        if not _dead:
            self._children = [ p.pid for p in proc.children() ]

        with proc.oneshot():
            if proc.is_running():
                self.rss = proc.memory_info().rss
                self.vms = proc.memory_info().vms
                self.ctx_vol = proc.num_ctx_switches().voluntary
                self.ctx_invol = proc.num_ctx_switches().involuntary
                self._cmdline = proc.cmdline()
                self.pcpu = None
                #self.pcpu += proc.cpu_percent(interval=DEFAULT_INTERVAL)
            else:
                self.rss = 0
                self.vms = 0
                self.ctx_vol = 0
                self.ctx_invol = 0
                self.cmdline = [ ]
                self.pcpu = 0.0

    def __str__(self):
        return "<{} name={} pid={} _pgid={} alive={} >\n".format(self.__class__.__name__,
                self.name(), self.pid,
                self._pgid, self.alive)

    def __repr__(self):
        return self.__str__()

    def __getattribute__(self, name: str):
        """
            Masquerade psutil.Process's fields and methods
        """
        if name.startswith("_"):
            return object.__getattribute__(self, name)
        elif name == "alive":
            return self._proc.is_running()
        elif name == "pidm":
            def __wrapper(x):
                try:
                    return self._pt.mirrors[x]
                except KeyError:
                    return False
            return __wrapper
        elif name == "parent":
            return lambda: self.pidm(self._parent)
        elif name == "children":
            return lambda: [ self.pidm(p) for p in self._children ]
        elif name == "cmdline":
            return lambda: self._cmdline
        elif hasattr(self._proc, name):
            return getattr(self._proc, name)
        else:
            return object.__getattribute__(self, name)

    def set_pcpu(self, value):
        """
            CPU percent values cannot be cached, so we need this
            trick to set it for all ProcessMirrors in one pass
        """
        self.pcpu = value

class EmptyBundle(Exception): pass

class ProcBundle:
    """
        Class represents bundle or group of processes associated
        by there process group ID and child-parent relations
    """
    def __init__(self, proclist, pt=None):

        self.proctable = pt

        if isinstance(proclist, ProcessMirror):
            self._pgid = proclist._pgid
            self.proclist = [ proclist ]

        if isinstance(proclist, list):
            self._pgid = proclist[0]._pgid
            self.proclist = proclist
            if len(proclist) == 0:
                raise EmptyBundle

        if isinstance(proclist, psutil.Process):
            self._pgid = os.getpgid(proclist.pid)
            self.proclist = [ proclist ]

        proc = self.proclist[0]
        self.leader = proc

        while utils.parent_has_single_child(proc):
            if not utils.is_kernel_thread(proc):
                proc = proc.parent()
                if proc and not utils.is_kernel_thread(proc) and proc not in self.proctable.bundled():
                    self.append(proc)
                    self.proctable._clear_expendable()
            else: break

        def go_top(proc):
            while proc and not utils.is_kernel_thread(proc):
                proc = proc.parent()
            return proc

        def go_bottom(proc):
            if not proc:
                return proc
            while not (proc._pgid == self._pgid or len(proc.children()) > 1):
                children = proc.children()
                if len(children) > 0:
                    proc = children[0]
                else:
                    return proc
            return proc

        lp = go_bottom(go_top(self.proclist[0]))


        if lp:
            self.leader = lp
            children_pids = [ p.pid for p in lp._proc.children(recursive=True) ]
            for p in children_pids:
                mirror = self.proctable.mirror_by_pid(p)
                if mirror not in self.proctable.bundled():
                    self.append(mirror)

        try:
            self.bundle_name = self._name_from_self_proclist()
        except:
            self.bundle_name = None

        if not self.bundle_name:
            self.bundle_name = self.name_from_cmdargs(self.leader)

    def append(self, proc: ProcessMirror):
        """
            Append ProcessMirror to curent ProcBundle's processes list
        """
        self.proclist.append(proc)

    def merge(self, bundle):
        """
            Merge ProcBundle with current one
        """
        self.proclist.extend(bundle.proclist)
        self.proctable.bundles.remove(bundle)

    def _name_from_self_proclist(self) -> str:
        """
            Generate bundle name from aquired proclist
        """
        out = ""
        out = list(map(lambda x: x._cmdline, self.proclist ))
        filter_titles = lambda x: len(x) > 1 and len(list(filter(lambda x: x == '', x))) > 0
        # filter out all processes except proctitles
        proctitles = [ x[0] for x in list(filter(filter_titles, out)) ]
        out = utils.reduce_sequence(list(filter(lambda x: not x.startswith("sshd"),
            proctitles)))
        return out

    @staticmethod
    def name_from_cmdargs(proc: ProcessMirror):
        """
            Choses name for the ProcBundle from ProcessMirror's
            command line arguments
        """
        _cmdline = proc._cmdline
        if utils.is_kernel_thread(proc): return "kernel"
        # special case for processes with set-up titles
        # due some specific behaviour of /proc they are have
        # empty strings in their cmdargs, for example:
        # [ 'zaggregator', '', '', '', '' ]
        # using this behaviour we are identifying such processes
        # and returning proctitle itself
        if len(_cmdline) > 1 and len(list(filter(lambda x: x == '', _cmdline))) > 0:
            return _cmdline[0]
        cline = list(filter(lambda x: not x.startswith("-"), _cmdline))

        def not_interpreter(word) -> bool:
            """ singleton for simplier parsing """
            for i in interpreters:
                if word.find(i) > -1:
                    return False
            return True

        if cline and cline[0].startswith("/"):
            cline[0] =  os.path.basename(cline[0])
        out = ":".join(filter(None,filter(not_interpreter, cline)))
        try:
            out = out.split()[0].strip(":")[:20]
        except IndexError:
            out = os.path.basename(_cmdline[0])

        return out

    def get_n_ctx_switches_vol(self) -> int:
        """
            Returns sum of voluntary context switches for all processes
            in the current bundle
        """
        return sum([p.ctx_vol for p in  self.proclist])

    def get_n_ctx_switches_invol(self) -> int:
        """
            Returns sum of involuntary context switches for all processes
            in the current bundle
        """
        return sum([p.ctx_invol for p in  self.proclist])

    def get_memory_info_rss(self) -> int:
        """
            Returns sum of resident memory sizes for all processes
            in the current bundle
        """
        return sum([p.rss for p in  self.proclist])

    def get_memory_info_vms(self) -> int:
        """
            Returns sum of virtual memory sizes for all processes
            in the current bundle
        """
        return sum([p.vms for p in  self.proclist])

    def get_cpu_percent(self) -> float:
        """
            Returns sum of consumed CPU percent for all processes
            in the current bundle
        """
        return sum([p.pcpu for p in  self.proclist])


def alive_or_false(proc):
    """ Singleton to filter-out dead processes """
    if proc.is_running():
        return proc
    return False

def process_call_guard(func, *args, **kwargs):
    """ Singleton to handle situation when process died
        after registering, but before sampling """

    try:
        return func(*args, **kwargs)
    except psutil.NoSuchProcess:
        return False

class ProcTable:
    """
        Class represents cached process table for the sampling period
    """
    def __init__(self):
        self._procs = []
        # fill the self._procs with ProcessMirrors of psutil processes
        _pcs = list(psutil.process_iter())
        list(map(self._procs.append, map(ProcessMirror, _pcs, len(_pcs)*[self] )))
        # set cpu_percent sampler
        [ process_call_guard(p.cpu_percent) for p in self._procs ]
        # wait for samples to collect
        time.sleep(DEFAULT_INTERVAL)
        # collect the samples
        pcpu = [ process_call_guard(p.cpu_percent) for p in self._procs ]
        # and map them to the corresponding ProcessMirror object
        list(map(lambda a,b: a.set_pcpu(b), self._procs, pcpu))
        # fill mirrors dict for faster mirror searching
        self.mirrors = {}
        for p in self._procs: self.mirrors.setdefault(p.pid, p)

        self.bundles = []

        # create expendable copy of the process list)
        self._expendable = list(self._procs)

        # collect the group id's of process groups
        pgids = list(set([p._pgid for p in self._procs]))

        #kernelProcs = list(filter(lambda x: utils.is_kernel_thread(x), self._expendable))
        #list(map(self._expendable.remove, kernelProcs))

        mirrors_by_pgid = list(map(self.mirrors_by_pgid, pgids))
        mirrors_by_pgid.sort(key=lambda x: len(x), reverse=True)
        while len(self._expendable) > 0:
            for ms in mirrors_by_pgid:
                    self.bundles.append(ProcBundle(ms, pt=self))
                    self._clear_expendable()

        uniq_names = list(set(self.get_bundle_names()))
        for name in uniq_names:
            n = self._get_bundles_by_name(name)
            for b in n[1:]:
                n[0].merge(b)

        #    self.get_bundle_names()



    def _clear_expendable(self):
        """ Clean up expendable processes list from bundled processes """
        list(map(lambda p: self._expendable.remove(p) if p in self.bundled() and p in
            self._expendable else None, self._expendable))


    def mirrors_by_pgid(self, pgid: int) -> [ProcessMirror]:
        """ Returns ProcessMirrors by Process Group ID (pgid) """
        return list(filter(lambda x: x._pgid == pgid,  self._procs))

    def mirror_by_pid(self, pid: int) -> ProcessMirror:
        """ Returns ProcessMirrors by process pid """
        if pid in self.mirrors.keys():
            return self.mirrors[pid]
        else:
            return None

    def bundled(self) -> list:
        """
            Returns list of all ProcessMirrors in all registered ProcBundles
        """
        ret = []
        for b in self.bundles:
            ret.extend(b.proclist)
        return ret

    def get_bundle_names(self) -> list:
        """
            Returns list of names of registered ProcBundles
        """
        return [ b.bundle_name for b in self.bundles ]

    def _get_bundles_by_name(self, name: str):
        """
            Returns list of all registered bundles with the same name. In the
            normal state there shouldn't be any duplicates in the regisered
            Procbundle, so we need to get all the names to merge ProcBundles
            with similar names. Should't be called outside of ProcTable internal
            context.
        """
        if name in self.get_bundle_names():
            return list(filter(lambda x: x.bundle_name == name, self.bundles))
        return None

    def get_bundle_by_name(self, name: str):
        """
            Returns ProcBundle with desired name or None for non-existing bundles
        """
        if name in self.get_bundle_names():
            return list(filter(lambda x: x.bundle_name == name, self.bundles))[0]
        return None

    def get_idle(self, interval=DEFAULT_INTERVAL):
        """
            TODO: refactor this
        """
        return psutil.cpu_times_percent(interval=interval).idle



import unittest
import psutil
import time
import logging
import random
import inspect
import signal
import os

import zaggregator
import zaggregator.utils as utils
import zaggregator.tests as tests
import zaggregator.procbundle as pb
from zaggregator.procbundle import ProcBundle, ProcTable, ProcessMirror
from zaggregator.tests import cycle

zaggregator.procbundle.DEFAULT_INTERVAL=0.1

class TestProcMirror(tests.TestCase):
    def test_ProcessMirror_alive(self):
        proc = psutil.Process(pid=os.getpid())
        mirror = ProcessMirror(proc, None)
        self.assertTrue(mirror.alive)
        self.assertTrue(mirror.alive == proc.is_running())

    def test_ProcessMirror_pid(self):
        proc = psutil.Process(pid=os.getpid())
        mirror = ProcessMirror(proc, None)
        self.assertTrue(mirror.pid == proc.pid == os.getpid())

    def test_ProcessMirror_set_pcpu(self):
        proc = psutil.Process(pid=os.getpid())
        mirror = ProcessMirror(proc, None)
        mirror.set_pcpu(0.123456)
        self.assertTrue(mirror.pcpu == 0.123456)

class TestProcTable(tests.TestCase):
    def test_ProcTable_mirror_by_pid(self):
        pt = ProcTable()
        proc = psutil.Process(pid=os.getpid())
        self.assertTrue(pt.mirror_by_pid(os.getpid()).pid == os.getpid() )

    def test_ProcTable_mirrors_by_pgid(self):
        bname = 'unittest-ptmbp'
        bunch, myproc, psutilproc = tests.BunchProto.start(bname)
        pt = ProcTable()
        pgid = os.getpgid(psutilproc.pid)
        self.assertTrue(
                len(
                    list(filter(None, [ p._pgid != pgid for p in pt.mirrors_by_pgid(pgid)]))
                    ) == 0
                )

        bunch.stop()

    def test_ProcTable_mirror_by_pid(self):
        bname = 'unittest-ptmbp2'
        bunch, myproc, psutilproc = tests.BunchProto.start(bname)
        pt = ProcTable()
        self.assertIsInstance(pt.mirror_by_pid(psutilproc.pid), ProcessMirror)

        bunch.stop()

    def test_ProcTable_pidm(self):
        bname = 'unittest-pm'
        bunch, myproc, psutilproc = tests.BunchProto.start(bname)
        pt = ProcTable()
        self.assertIsInstance(pt.mirror_by_pid(psutilproc.pid).pidm(psutilproc.pid), ProcessMirror)

        bunch.stop()

    def test_ProcTable_children(self):
        bname = 'unittest-pm'
        nchildren = 5
        bunch, myproc, psutilproc = tests.BunchProto.start(bname, nchildren=5)
        pt = ProcTable()
        m = pt.mirror_by_pid(psutilproc.pid)
        self.assertTrue(len(m.children()) == nchildren)
        self.assertIsInstance(m.children()[0], ProcessMirror)

        bunch.stop()

    def test_ProcTable_test(self):
        bname = 'unittest-t'
        nchildren = 5
        bunch, myproc, psutilproc = tests.BunchProto.start(bname, nchildren=5)
        pt = ProcTable()

        #print(pt.bundles[0].bundle_name)

        bunch.stop()

if __name__ == '__main__':
    tests.run_test_module_by_name(__file__)



import sqlite3

DBPATH="/var/run/zaggregator/zaggregator.sqlite"
db = None

def __init__() -> None:
    """ Initalize sqlite dataabase """
    global db
    db = sqlite3.connect(DBPATH)
    create_table_str = """
        CREATE TABLE
        IF NOT EXISTS
        samples (
            ts DATETIME DEFAULT CURRENT_TIMESTAMP,
            name TEXT,
            memrss INT,
            memvms INT,
            ctxvol INT,
            ctxinvol INT,
            pcpu REAL
        );
        """;
    create_trigger = """
        CREATE TRIGGER IF NOT EXISTS
        DELETE_TAIL
        AFTER INSERT ON samples
        BEGIN
            DELETE FROM samples WHERE ts not in (select ts from samples order by ts desc
            limit 300);
        END;
    """
    db.execute(create_table_str)
    db.execute(create_trigger)

class BadRecord(Exception): pass

def add_record(record) -> None:
    """
        Add record into sqlite database
    """
    if len(record) != 3 and hasattr(record, "__iter__"):
        query = """
            INSERT INTO samples
            ( name, memrss, memvms, ctxvol, ctxinvol, pcpu )
            VALUES ('{}',{});""".format(record[0], ",".join(map(str, record[1:])))

        db.execute(query)
        db.commit()
    else:
        raise BadRecord

def get_bundle_names() -> [str]:
    """
        Get list of bundle names from sqlite database
    """
    query = """
        SELECT DISTINCT(name) FROM samples;
        """
    return [ row[0] for row in db.execute(query) ]

def get(bname:str, check:str):
    """
        Get value of `check' variable for bundle with name `bname'
    """
    query = """
        SELECT {} FROM samples
        WHERE name='{}' AND
            ( (julianday('now') - julianday(ts))*24*60*60 < 30 )
        ORDER BY ts DESC
        LIMIT 1;
        """.format(check, bname)
    result = list(db.execute(query))
    if len(result) > 0:
        return result[0][0]

__init__()



import unittest
from fuzzywuzzy import fuzz, StringMatcher
import psutil
import os, sys
from multiprocessing import Process
import time
import random
import inspect
import logging

import zaggregator.utils as utils
import zaggregator.tests as tests
from zaggregator.procbundle import ProcBundle, ProcessMirror
from zaggregator.procbundle import ProcTable as ProcessTable

class TestZaggregatorUtils(tests.TestCase):

    fuzzy_string_sets = ([
        "/usr/sbin/zabbix_agentd -c /etc/zabbix/zabbix_agentd.conf",
        "/usr/sbin/zabbix_agentd: collector [idle 1 sec]",
        "/usr/sbin/zabbix_agentd: listener #1 [waiting for connection]",
        "/usr/sbin/zabbix_agentd: listener #2 [waiting for connection]",
        "/usr/sbin/zabbix_agentd: listener #3 [waiting for connection]",
        "/usr/sbin/zabbix_agentd: active checks #1 [idle 1 sec]",
        ],
        [
            "postgres: checkpointer process",
            "postgres: writer process",
            "postgres: wal writer process",
            "postgres: autovacuum launcher process",
            "postgres: stats collector process",
        ],
        [
            "php-fpm: pool main",
            "php-fpm: pool main",
            "php-fpm: pool main",
            "php-fpm: pool main",
            ],
        )

    fuzzy_string_sets_nomatch = ([
        "/usr/sbin/zabbix_agentd: active checks #1 [idle 1 sec]",
        "postgres: checkpointer process",
        "php-fpm: pool main",
        ],
        [ 'YK1UYU2C: child#0',
          'LCAR8XLT: child#1',
          '6I6P9FKE: child#2',
          '4ZTJ1YEH: child#3',
          '1Y261N7W: child#4',
          'unittest: master',],
        )


    def test_fuzzy_match_seq(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        for i in self.fuzzy_string_sets:
            self.assertTrue(utils.fuzzy_sequence_match(i))
        for i in self.fuzzy_string_sets_nomatch:
            self.assertFalse(utils.fuzzy_sequence_match(i))

    def test_fuzzy_match(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        for fuzzy_strings in self.fuzzy_string_sets:
            for i in range(len(fuzzy_strings)-1):
                self.assertTrue(utils.fuzzy_match(*fuzzy_strings[i:i+2]))

    def test_fuzzy_namesearch(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        r = []
        for s in self.fuzzy_string_sets:
            r.append(utils.reduce_sequence(s))

        r1 = [
            '/usr/sbin/zabbix_agentd',
            'postgres',
            'php-fpm: pool main']

        self.assertTrue(r == r1)

    def test_is_proc_group_parent(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittests"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname)

        self.assertTrue(utils.is_proc_group_parent(psutilproc))

        bunch.stop()

    def test_proc_is_not_group_parent(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittests-np"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname, israndom=True)

        self.assertFalse(utils.is_proc_group_parent(psutilproc))

        bunch.stop()

    def test_parent_has_single_child(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittest-sc"
        bunch, myproc, psutilproc = tests.bunchproto.start(bname,
                israndom=random.choice((true,false)), nchildren=1)

        self.assertTrue(utils.parent_has_single_child(psutilproc.children()[0]))

        bunch.stop()

    def test_parent_has_single_child(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittest-nsc"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname,
                israndom=random.choice((True,False)), nchildren=2)

        self.assertFalse(utils.parent_has_single_child(psutilproc.children()[0]))

        bunch.stop()

    def test_parent_has_single_child_false(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittest-nscf"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname,
                israndom=random.choice((True,False)), nchildren=1)
        ProcessMirror,
        self.assertTrue(utils.parent_has_single_child(psutilproc.children()[0]))

        bunch.stop()

    def test_parent_has_single_child_init(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        # by init I meant any top-level process, it could be systemd, launchctl and so on
        psutilproc = random.choice(psutil.Process(pid=1).children())
        self.assertFalse(utils.parent_has_single_child(psutilproc))

    def test_parent_has_single_child_chain(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittest-ch"
        nproc = random.choice([i for i in range(2,10)])
        bunch, myproc, psutilproc = tests.BunchProto.start(bname, nchildren=nproc)

        pt = ProcessTable()
        bundle = ProcBundle(ProcessMirror(psutilproc, pt), pt=pt)
        #self.assertTrue(len(bundle.proclist)+1 > nproc)

        bunch.stop()

    def test_is_kernel_thread_false(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittest-kth"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname,
                israndom=random.choice((True,False)), nchildren=1)

        self.assertFalse(utils.is_kernel_thread(psutilproc))

        bunch.stop()

    def test_is_kernel_thread_true(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        pt = ProcessTable()
        proc = ProcessMirror(psutil.Process(pid=2), pt)
        self.assertTrue(utils.is_kernel_thread(proc))

    def test_is_leaf_process_true(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittest-lpt"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname,
                israndom=random.choice((True,False)), nchildren=1)

        self.assertTrue(utils.is_leaf_process(psutilproc.children()[0]))

        bunch.stop()

    def test_is_leaf_process_false(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittest-lpf"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname,
                israndom=random.choice((True,False)), nchildren=2)

        self.assertFalse(utils.is_leaf_process(psutilproc))
        self.assertFalse(utils.is_leaf_process(psutilproc.children()[0]))

        bunch.stop()

    def test_is_proc_in_bundle(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = 'unittest-ipib'
        bunch, myproc, psutilproc = tests.BunchProto.start(bname)
        pt = ProcessTable()
        bundle = ProcBundle(ProcessMirror(psutilproc, pt), pt=pt)
        proc = psutilproc.children()[0]

        #self.assertTrue(utils.is_proc_in_bundle(proc, bundle))

        bunch.stop()

    def test_is_proc_in_bundle_false(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = '1-unittest-ipibf'
        (bunch, myproc, psutilproc) = tests.BunchProto.start(bname)
        bname = '2-unittest-ipibf'
        (bunch2, myproc2, psutilproc2) = tests.BunchProto.start(bname)

        pt = ProcessTable()
        bundle = ProcBundle(ProcessMirror(psutilproc, pt), pt=pt)
        proc = psutilproc.children()[0]

        self.assertFalse(utils.is_proc_in_bundle(myproc2, bundle))

        bunch.stop()
        bunch2.stop()


    """
    def test_proc_similar_to(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = 'unittest-ipst'
        bunch, myproc, psutilproc = tests.BunchProto.start(bname)
        bundle = ProcBundle(ProcessMirror(psutilproc, None))

        self.assertTrue(utils.is_proc_similar_to(*bundle.proclist[1:3]))

        bunch.stop()


    def test_proc_similar_to(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = 'unittest-ipst'
        bunch, myproc, psutilproc = tests.BunchProto.start(bname)
        bundle = ProcBundle(ProcessMirror(psutilproc, None))

        #self.assertTrue(utils.is_proc_similar_to(*bundle.proclist[1:3]))

        bunch.stop()

    def test_ProcessGone_chain(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = "unittest-pgch"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname, nchildren=4)

        def term(self):
            self.leader[0].send_signal(9)
        ProcBundle._collect_chain_hook = term
        try:
            bundle = ProcBundle(psutilproc)
        except Exception as e:
            print(e)
        """

if __name__ == '__main__':
    run_test_module_by_name(__file__)



import asyncio
import sys, os
import zaggregator
import time
import setproctitle
from zaggregator import sqlite


if len(sys.argv) > 1:
    pidfile = sys.argv[1]
    with open(pidfile, "w") as fd:
        fd.write(str(os.getpid()))

delay = 30
loop = asyncio.get_event_loop()

def collect_data(bundle) -> (str, int, int, float):
    """  Collect data """
    return  (bundle.bundle_name, bundle.get_n_ctx_switches_vol(),
            bundle.get_memory_info_rss(),
            bundle.get_cpu_percent())

def zag_sampler_loop(lc):
    """ Main sampler loop """
    loop, callback = lc

    loop.call_later(delay, callback, lc)
    pt = zaggregator.ProcTable()
    for n in pt.get_bundle_names():
        b = pt.get_bundle_by_name(n)
        sqlite.add_record(
                (
                    b.bundle_name,
                    b.get_memory_info_rss(),
                    b.get_memory_info_vms(),
                    b.get_n_ctx_switches_vol(),
                    b.get_n_ctx_switches_invol(),
                    b.get_cpu_percent(),
                    ))

def start() -> None:
    """
        initialize and start main daemon loop
    """

    setproctitle.setproctitle('zaggregator')
    lc = (loop, callback)
    loop.call_later(delay, callback, lc)

    loop.run_forever()
    loop.close()

callback = zag_sampler_loop



import unittest
import os
import logging as log
import random
import psutil
import string
import os, time
import sys
from multiprocessing import Process

VERBOSITY = 2
HERE = os.path.abspath(os.path.dirname(__file__))

class TestCase(unittest.TestCase):

    # Print a full path representation of the single unit tests
    # being run.
    def __str__(self):
        return "%s.%s.%s" % (
            self.__class__.__module__, self.__class__.__name__,
            self._testMethodName)

    # assertRaisesRegexp renamed to assertRaisesRegex in 3.3;
    # add support for the new name.
    if not hasattr(unittest.TestCase, 'assertRaisesRegex'):
        assertRaisesRegex = unittest.TestCase.assertRaisesRegexp

# override default unittest.TestCase
unittest.TestCase = TestCase

LOGFILE="test.log"
log.basicConfig(filename=LOGFILE, level=log.DEBUG)
print("Unittest log stored at {}".format(os.path.realpath(LOGFILE)))

class BunchProto:

    sleeptime = 5.0
    ttime = 0.1

    def __init__(self, name="defname", israndom=False, nchildren=5, rstrlen=8, func=None):
        self.nchildren = nchildren
        self.rstrlen = rstrlen
        self.name = name
        self.israndom = israndom
        self.master = Process(target=self.process_proto_parent)
        self.psutilproc = None
        self.func = func

    def name_or_random(self, i):
        name = self.name
        if self.israndom:
            name = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in
                    range(self.rstrlen))
        return "{}: child#{!s}".format(name, i)

    @staticmethod
    def process_proto(name, sleeptime, func):
        import setproctitle
        setproctitle.setproctitle(name)

        if func or not sleeptime:
            func()
        else:
            time.sleep(BunchProto.sleeptime)

    def process_proto_parent(self):
        import setproctitle
        import signal
        import sys

        def sig_mitigator(signum, frame):
            def _stop(p):
                class fake_fd:
                    def terminate(self):
                        pass
                if p._popen == None:
                    p._popen = fake_fd()
                p.terminate()
            [ _stop(p) for p in self.pcs ]
            sys.exit(0)

        signal.signal(signal.SIGTERM, sig_mitigator)

        setproctitle.setproctitle("{}: master".format(self.name))
        self.pcs = [ Process(target=BunchProto.process_proto,
            args=(self.name_or_random(i), self.sleeptime, self.func)) for i in range(self.nchildren) ]
        [ p.start() for p in self.pcs ]

    @staticmethod
    def start(name, israndom=False, nchildren=5, rstrlen=8, func=None ):
        bunch = BunchProto(name, israndom=israndom, nchildren=nchildren, rstrlen=rstrlen,
                func=func)
        myproc = bunch.master
        myproc.start()
        time.sleep(BunchProto.ttime) # give process a moment to set it's title properly
        psutilproc = psutil.Process(pid=myproc.pid)
        bunch.psutilproc = psutilproc
        return bunch,myproc,psutilproc

    def stop(self):
        def _stop(p):
            if p.is_running():
                p.terminate()
        [ _stop(p) for p in psutil.Process(pid=self.master.pid).children() ]
        if psutil.Process(pid=self.master.pid).is_running():
            self.master.terminate()

def cycle():
    # singleton, which does nothing, only consumes CPU
    import signal
    import sys

    def sig_mitigator(signum, frame):
        sys.exit(0)

    # define and install custom SIGTERM handler
    signal.signal(signal.SIGTERM, sig_mitigator)

    while True:
        pass

def get_suite():
    testmods = [os.path.splitext(x)[0] for x in os.listdir(HERE)
                if x.endswith('.py') and x.startswith('test_') and not
                x.startswith('test_memory_leaks')]
    if "WHEELHOUSE_UPLOADER_USERNAME" in os.environ:
        testmods = [x for x in testmods if not x.endswith((
                    "osx", "posix", "linux"))]
    suite = unittest.TestSuite()
    for tm in testmods:
        # ...so that the full test paths are printed on screen
        tm = "zaggregator.tests.%s" % tm
        suite.addTest(unittest.defaultTestLoader.loadTestsFromName(tm))
    return suite

def run_suite():
    _setup_tests()
    result = unittest.TextTestRunner(verbosity=VERBOSITY).run(get_suite())
    success = result.wasSuccessful()
    sys.exit(0 if success else 1)

def _setup_tests():
    pass

def run_test_module_by_name(name):
    testmodules = [os.path.splitext(x)[0] for x in os.listdir(HERE)
                    if x.endswith('.py') and x.startswith('test_')]
    _setup_tests()
    name = os.path.splitext(os.path.basename(name))[0]
    suite = unittest.TestSuite()
    suite.addTest(unittest.defaultTestLoader.loadTestsFromName(name))
    result = unittest.TextTestRunner(verbosity=VERBOSITY).run(suite)
    success = result.wasSuccessful()
    sys.exit(0 if success else 1)



import unittest
import psutil
import time
import logging
import random
import inspect
import signal

import zaggregator
import zaggregator.utils as utils
import zaggregator.tests as tests
import zaggregator.procbundle as pb
from zaggregator.procbundle import ProcBundle, ProcTable, ProcessMirror
from zaggregator.tests import cycle
zaggregator.procbundle.DEFAULT_INTERVAL=0.1

class TestProcBundle(tests.TestCase):

    def test_ProcessBundle_name(self):
        bname = 'unittest-pb'
        bunch, myproc, psutilproc = tests.BunchProto.start(bname)

        procs = psutilproc.children()
        procs.append(psutilproc)
        pt = ProcTable()
        procs = [ ProcessMirror(p,pt) for p in procs ]
        bundle = ProcBundle(procs, pt=pt)
        print(bundle.bundle_name)
        self.assertTrue(bundle.bundle_name == bname)

        bunch.stop()

    def test_ProcTable_get_bundle_names(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        try:
            table = ProcTable()
            logging.debug("Bundle names: {}".format(table.get_bundle_names()))
        except psutil._exceptions.AccessDenied as e:
            logging.error(e)
            logging.error("Some tests require root priveleges")

    """

    def test_ProcBundle_merge(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = 'unittest-pbm'
        ((bunch1, myproc1, psutilproc1),(bunch2, myproc2, psutilproc2), (bunch3, myproc3, psutilproc3)) = (tests.BunchProto.start(bname),
                tests.BunchProto.start(bname),
                tests.BunchProto.start(bname))

        pt = ProcTable()
        bundle1,bundle2,bundle3 = ( ProcBundle([ProcessMirror(psutilproc1, pt)]),
                ProcBundle([ProcessMirror(psutilproc2, pt)]),
                ProcBundle([ProcessMirror(psutilproc3, pt)]) )
        logging.debug(len(bundle1.proclist))
        bundle1.merge(bundle2)
        bundle1.merge(bundle3)
        logging.debug(len(bundle1.proclist))


        self.assertTrue(utils.is_proc_in_bundle(bundle2.proclist[1], bundle1))
        self.assertTrue(bundle2.leader[0] in  bundle2.leader)

        bunch1.stop()
        bunch2.stop()
        bunch3.stop()

    """
    def test_ProcBundle_stats(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        bname = 'unittest-pbs'
        bunch, myproc, psutilproc = tests.BunchProto.start(bname)

        pt = ProcTable()
        bundle = pt.get_bundle_by_name(pt.get_bundle_names()[0])
        self.assertIsInstance(bundle.get_n_ctx_switches_vol(), int)
        self.assertIsInstance(bundle.get_n_ctx_switches_invol(), int)
        self.assertIsInstance(bundle.get_memory_info_rss(), int)
        self.assertIsInstance(bundle.get_memory_info_vms(), int)
        self.assertIsInstance(bundle.get_cpu_percent(), float)

        bunch.stop()

    def test_get_idle(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        p = ProcTable()
        self.assertIsInstance(p.get_idle(), float)

        bname = "unittest-gi"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname, nchildren=2, func=cycle)
        p = ProcTable()

        self.assertTrue(p.get_idle() < 10.0)

        bunch.stop()

    def test_get_pcpu_busy(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])

        bname = "unittest-gpcpb"
        procname = "sh:./test.sh"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname, nchildren=2, func=cycle)
        p = ProcTable()
        time.sleep(0.1)
        #print(p.get_bundle_names())
        """
        for b in p.bundles:
            print("{}:\t{}".format(b.bundle_name, b.proclist))
        """
        bundle = p.get_bundle_by_name(bname)
        pcpu = bundle.get_cpu_percent()
        #print(bundle.__class__)
        pcpu_threshold = 75
        if pcpu <= pcpu_threshold:
            print("pcpu value: {}".format(pcpu))
        self.assertTrue(pcpu > pcpu_threshold)


        bunch.stop()

    def test_name_from_proctitles(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])

        bname = "unittest-nfpt"
        procname = "sh:./test.sh"
        bunch, myproc, psutilproc = tests.BunchProto.start(bname, nchildren=2, func=cycle)
        p = ProcTable()
        time.sleep(0.1)
        bundle = p.get_bundle_by_name(bname)
        print(bundle._name_from_self_proclist())
        bunch.stop()
        """
    """

if __name__ == '__main__':
    run_test_module_by_name(__file__)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zaggregator",
    version="0.0.4",
    author="Denis Barov",
    author_email="dindin+zaggregator@dindin.ru",
    description="Per-process stat monitoring solution for Zabbix",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fordindin/zaggregator",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    keywords="zabbix monitoring",
    python_requires='>=3',
    install_requires=['fuzzywuzzy>=0.16.0', 'psutil>=5.4.5', 'python-Levenshtein>=0.12.0', 'setproctitle>=1.1.10'],
    packages=setuptools.find_packages(exclude=['contrib', 'docs', 'tests*', 'venv', 'misc', 'build', 'dist']),
    zip_safe=True,
    entry_points={
        'console_scripts': [
            'zaggregator = zaggregator.daemon:start',
            'zcheck = zaggregator.client:main',
        ],},
    data_files=[
        ('/var/run/zaggregator', []),
        ('/usr/share/zaggregator', ['misc/zaggregator.service', 'misc/zaggregator.conf', 'README.md', 'LICENSE',]),
        ('/usr/share/zaggregator/init.d', ['misc/init.d/zaggregator'],),
            ],

)



#!/usr/bin/env python

from difflib import SequenceMatcher
from fuzzywuzzy import fuzz, StringMatcher
import os, sys
import logging as log
import psutil
import json
import zaggregator

DEFAULT_FUZZY_THRESHOLD = 53
class ProcessGone(Exception): pass

def eprint(*args, **kwargs):
    """
        Wrapper around built-in `print' function which allow to print
        into sys.stderr instead of sys.stdout
    """
    print(*args, file=sys.stderr, **kwargs)

def reduce_sequence(seq:list) -> str:
    """
        Reduce sequence of names to single name
    """
    if len(seq) == 1: return seq[0]

    def single_pass(iseq):
        matches = []
        fuzzy_strings = iseq
        for i in range(len(fuzzy_strings)-1):
            a = fuzzy_strings[i]
            b = fuzzy_strings[i+1]
            sm = SequenceMatcher(a=a, b=b)
            match = sm.get_matching_blocks()[0]
            matches.append(a[match.a:match.size])
        return matches

    sp = single_pass(seq)

    while len(sp) > 1:
        sp = single_pass(sp)

    return sp[0].strip("[]:- ")

def fuzzy_match(a:str, b:str, threshold=DEFAULT_FUZZY_THRESHOLD) -> bool:
    """
        Checks if there a fussy match between two strings
    """
    score = fuzz.partial_ratio(a,b)
    log.debug("Fuzzy score: {} ({},{})".format(score, a, b))
    if score > threshold:
        return True
    else:
        return False

def fuzzy_sequence_match(seq:list) -> bool:
    """
        Checks if all strings in sequence are fuzzy match
    """
    for i in range(len(seq)-1):
        match = fuzzy_match(seq[i],seq[i+1])
        if not match:
            return False
    return True

def is_proc_group_parent(proc) -> bool:
    """
        Checks if process is a group parent
    """
    if os.uname().sysname == 'Darwin':
        fproc_names = filter(lambda x: len(x.cmdline()) > 0, proc.children())
        procs_names = [ p.cmdline()[0] for p in fproc_names ]
        procs_names.append(proc.cmdline()[0])
    else:
        procs_names = [ p.name() for p in proc.children() ]
        procs_names.append(proc.name())
    if len(procs_names) <= 1:
        return False
    if fuzzy_sequence_match(procs_names):
        return True
    #if len(proc.children()) > 2 and proc.parent():
    #    return True
    return False

def parent_has_single_child(proc) -> bool:
    """
        Checks if process's parent has only one child
    """
    if not proc or not proc.parent: return False
    try:
        parent = proc.parent()
        if not parent: return False
        return len(proc.parent().children()) == 1
    except psutil.NoSuchProcess:
        raise ProcessGone


def is_kernel_thread(proc) -> bool:
    """
        Checks if process is a kernel process/thread
    """
    if isinstance(proc, psutil.Process):
        if os.getpgid(proc.pid) == 0:
            return True
    elif isinstance(proc, zaggregator.ProcessMirror):
        if proc._pgid == 0:
            return True
    return False

def is_leaf_process(proc) -> bool:
    """
        Checks if process has no children
    """

    try:
        if len(proc.children()) == 0 and len(proc.parent().children()) == 1:
                return True
        return False
    except psutil.NoSuchProcess:
        raise ProcessGone

def is_proc_in_bundle(proc, bundle) -> bool:
    """
        Checks if process in ProcBundle
    """
    return (proc in bundle.proclist)

def is_proc_similar_to(proc1, proc2) -> bool:
    """
        Checks if two processes are similar
    """
    if fuzzy_match(proc1.cmdline(),proc2.cmdline(), threshold=90) \
            and proc1.cwd() == proc2.cwd():
                return True
    return False

def discovery_json(names):
    """
        Returns bundle names in Zabbix autodiscovery JSON format
    """
    template = { "data" : [ ]}
    for bn in names:
        template["data"].append({ "{#PROCGROUP}": bn, })
    template["data"].append({ "{#PROCGROUP}": 'idle', })
    return json.dumps(template)




#!/usr/bin/env python

# Copyright (c) 2009, Giampaolo Rodola'. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""
Run unit tests. This is invoked by:

$ python -m psutil.tests
"""

import contextlib
import optparse
import os
import sys

from zaggregator.tests import run_suite


HERE = os.path.abspath(os.path.dirname(__file__))

def main():
    usage = "%s -m psutil.tests [opts]" % sys.argv[0]
    parser = optparse.OptionParser(usage=usage, description="run unit tests")
    parser.add_option("-i", "--install-deps",
                      action="store_true", default=False,
                      help="don't print status messages to stdout")

    opts, args = parser.parse_args()
    run_suite()

main()



import unittest
import logging
import inspect
import sqlite3

import zaggregator.sqlite as sqlite
sqlite.DBPATH = ":memory:"
sqlite.__init__()
import zaggregator.tests as tests


class TestSqliteModule(tests.TestCase):

    records = (
            ("test0", 1, 2, 10, 41, 10.0),
            ("test1", 5, 2, 11, 42, 11.0),
            ("test2", 6, 8, 13, 43, 12.0),
            ("test3", 7, 9, 12, 44, 13.0),
            )

    def test_db(self):
        logging.debug("======= %s ======" % inspect.stack()[0][3])
        self.assertTrue(type(sqlite.db) == sqlite3.Connection)

    def test_add_get_record(self):
        for r in self.records:
            sqlite.add_record(r)

    def test_get_bundle_names(self):
        names = sqlite.get_bundle_names()
        self.assertTrue(names == [ r[0] for r in self.records ])


if __name__ == '__main__':
    run_test_module_by_name(__file__)



import argparse
import sys
from zaggregator import sqlite
from zaggregator.utils import discovery_json, eprint

checks = [ "pcpu", "memrss", "memvms", "ctxvol", "ctxinvol" ]


def discover():
    """ Returns bundles list in Zabbix autodiscovery JSON format """
    print(discovery_json(sqlite.get_bundle_names()))

def check(opts):
    """ Returns value for specified bundle and check type """
    bname, check = opts
    if check not in checks:
        eprint(
            "\tInvalid check argument: \'{}\'\n\tSupported options are: \'{}\'".format(
            check, "','".join(checks)))
        sys.exit(1)
    print(sqlite.get(bname, check))

def main(*args, **kwargs):
    """ main module """
    parser = argparse.ArgumentParser(description='Zabbix aggregator client.')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-discover', action='store_true',
        help="Discover and print out list of bundles in Zabbix autodiscovery JSON format.")
    group.add_argument('-bundle', nargs=2, metavar=("<bundleName>", "<check>"),
        help="Bundle name to check stats on. Check can be one of: pcpu, rss, vms, ctxvol, ctxinvol",)
    args = parser.parse_args()

    if args.discover:
        discover()
    if args.bundle:
        check(args.bundle)


if __name__ == '__main__':
    main()
